import os

import numpy as np
import pytest
from rustworkx import NoPathFound

from semantic_world.adapters.urdf import URDFParser
from semantic_world.connections import OmniDrive
from semantic_world.ik_solver import MaxIterationsException, UnreachableException
from semantic_world.prefixed_name import PrefixedName
from semantic_world.robots import PR2
from semantic_world.spatial_types.derivatives import Derivatives
from semantic_world.spatial_types.symbol_manager import symbol_manager
from semantic_world.world import World, Body
from semantic_world.testing import pr2_world


def test_compute_chain_of_bodies_pr2(pr2_world):
    root_link = pr2_world.get_body_by_name('base_footprint')
    tip_link = pr2_world.get_body_by_name('r_gripper_tool_frame')
    real = pr2_world.compute_chain_of_bodies(root=root_link, tip=tip_link)
    real = [x.name for x in real]
    assert real == [PrefixedName(name='base_footprint', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='base_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='torso_lift_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_shoulder_pan_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_shoulder_lift_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_upper_arm_roll_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_upper_arm_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_elbow_flex_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_forearm_roll_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_forearm_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_wrist_flex_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_wrist_roll_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_gripper_palm_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_gripper_tool_frame', prefix='pr2_kinematic_tree')]


def test_compute_chain_of_connections_pr2(pr2_world):
    root_link = pr2_world.get_body_by_name('base_footprint')
    tip_link = pr2_world.get_body_by_name('r_gripper_tool_frame')
    real = pr2_world.compute_chain_of_connections(root=root_link, tip=tip_link)
    real = [x.name for x in real]
    assert real == [PrefixedName(name='base_footprint_T_base_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='base_link_T_torso_lift_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='torso_lift_link_T_r_shoulder_pan_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_shoulder_pan_link_T_r_shoulder_lift_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_shoulder_lift_link_T_r_upper_arm_roll_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_upper_arm_roll_link_T_r_upper_arm_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_upper_arm_link_T_r_elbow_flex_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_elbow_flex_link_T_r_forearm_roll_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_forearm_roll_link_T_r_forearm_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_forearm_link_T_r_wrist_flex_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_wrist_flex_link_T_r_wrist_roll_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_wrist_roll_link_T_r_gripper_palm_link', prefix='pr2_kinematic_tree'),
                    PrefixedName(name='r_gripper_palm_link_T_r_gripper_tool_frame', prefix='pr2_kinematic_tree')]


def test_compute_chain_of_bodies_error_pr2(pr2_world):
    root = pr2_world.get_body_by_name('r_gripper_tool_frame')
    tip = pr2_world.get_body_by_name('base_footprint')
    with pytest.raises(NoPathFound):
        pr2_world.compute_chain_of_bodies(root, tip)


def test_compute_chain_of_connections_error_pr2(pr2_world):
    root = pr2_world.get_body_by_name('r_gripper_tool_frame')
    tip = pr2_world.get_body_by_name('base_footprint')
    with pytest.raises(NoPathFound):
        pr2_world.compute_chain_of_connections(root, tip)


def test_compute_split_chain_of_bodies_pr2(pr2_world):
    root = pr2_world.get_body_by_name('l_gripper_r_finger_tip_link')
    tip = pr2_world.get_body_by_name('l_gripper_l_finger_tip_link')
    chain1, connection, chain2 = pr2_world.compute_split_chain_of_bodies(root, tip)
    chain1 = [n.name.name for n in chain1]
    connection = [n.name.name for n in connection]
    chain2 = [n.name.name for n in chain2]
    assert chain1 == ['l_gripper_r_finger_tip_link', 'l_gripper_r_finger_link', ]
    assert connection == ['l_gripper_palm_link']
    assert chain2 == ['l_gripper_l_finger_link', 'l_gripper_l_finger_tip_link']


def test_get_split_chain_pr2(pr2_world):
    root = pr2_world.get_body_by_name('l_gripper_r_finger_tip_link')
    tip = pr2_world.get_body_by_name('l_gripper_l_finger_tip_link')
    chain1, chain2 = pr2_world.compute_split_chain_of_connections(root, tip)
    chain1 = [n.name.name for n in chain1]
    chain2 = [n.name.name for n in chain2]
    assert chain1 == ['l_gripper_r_finger_link_T_l_gripper_r_finger_tip_link',
                      'l_gripper_palm_link_T_l_gripper_r_finger_link']
    assert chain2 == ['l_gripper_palm_link_T_l_gripper_l_finger_link',
                      'l_gripper_l_finger_link_T_l_gripper_l_finger_tip_link']


def test_compute_fk_np_pr2(pr2_world):
    tip = pr2_world.get_body_by_name('r_gripper_tool_frame')
    root = pr2_world.get_body_by_name('l_gripper_tool_frame')
    fk = pr2_world.compute_forward_kinematics_np(root, tip)
    np.testing.assert_array_almost_equal(fk, np.array(
        [[1.0, 0.0, 0.0, -0.0356], [0, 1.0, 0.0, -0.376], [0, 0.0, 1.0, 0.0], [0.0, 0.0, 0.0, 1.0]]))


def test_compute_fk_np_l_elbow_flex_joint_pr2(pr2_world):
    tip = pr2_world.get_body_by_name('l_elbow_flex_link')
    root = pr2_world.get_body_by_name('l_upper_arm_link')

    fk_expr = pr2_world.compose_forward_kinematics_expression(root, tip)
    fk_expr_compiled = fk_expr.compile()
    fk2 = fk_expr_compiled.fast_call(symbol_manager.resolve_symbols(*fk_expr_compiled.symbol_parameters))

    np.testing.assert_array_almost_equal(fk2, np.array(
        [[0.988771, 0., -0.149438, 0.4], [0., 1., 0., 0.], [0.149438, 0., 0.988771, 0.], [0., 0., 0., 1.]]))

def test_compute_ik(pr2_world):
    bf = pr2_world.root
    eef = pr2_world.get_body_by_name('r_gripper_tool_frame')
    fk = pr2_world.compute_forward_kinematics_np(bf, eef)
    fk[0, 3] -= 0.2
    joint_state = pr2_world.compute_inverse_kinematics(bf, eef, fk)
    for joint, state in joint_state.items():
        pr2_world.state[joint.name].position = state
    pr2_world.notify_state_change()
    actual_fk = pr2_world.compute_forward_kinematics_np(bf, eef)
    assert np.allclose(actual_fk, fk, atol=1e-3)

def test_compute_ik_max_iter(pr2_world):
    bf = pr2_world.root
    eef = pr2_world.get_body_by_name('r_gripper_tool_frame')
    fk = pr2_world.compute_forward_kinematics_np(bf, eef)
    fk[2, 3] = 10
    with pytest.raises(MaxIterationsException):
        pr2_world.compute_inverse_kinematics(bf, eef, fk)

def test_compute_ik_unreachable(pr2_world):
    bf = pr2_world.root
    eef = pr2_world.get_body_by_name('base_footprint')
    fk = pr2_world.compute_forward_kinematics_np(bf, eef)
    fk[2, 3] = -1
    with pytest.raises(UnreachableException):
        pr2_world.compute_inverse_kinematics(bf, eef, fk)

def test_apply_control_commands_omni_drive_pr2(pr2_world):
    omni_drive: OmniDrive = pr2_world.get_connection_by_name('odom_combined_T_base_footprint')
    cmd = np.zeros((len(pr2_world.degrees_of_freedom)), dtype=float)
    cmd[pr2_world.state._index[omni_drive.x_vel.name]] = 100
    cmd[pr2_world.state._index[omni_drive.y_vel.name]] = 100
    cmd[pr2_world.state._index[omni_drive.yaw.name]] = 100
    dt = 0.1
    pr2_world.apply_control_commands(cmd, dt, Derivatives.jerk)
    assert pr2_world.state[omni_drive.yaw.name].jerk == 100.
    assert pr2_world.state[omni_drive.yaw.name].acceleration == 100. * dt
    assert pr2_world.state[omni_drive.yaw.name].velocity == 100. * dt * dt
    assert pr2_world.state[omni_drive.yaw.name].position == 100. * dt * dt * dt

    assert pr2_world.state[omni_drive.x_vel.name].jerk == 100.
    assert pr2_world.state[omni_drive.x_vel.name].acceleration == 100. * dt
    assert pr2_world.state[omni_drive.x_vel.name].velocity == 100. * dt * dt
    assert pr2_world.state[omni_drive.x_vel.name].position == 0

    assert pr2_world.state[omni_drive.y_vel.name].jerk == 100.
    assert pr2_world.state[omni_drive.y_vel.name].acceleration == 100. * dt
    assert pr2_world.state[omni_drive.y_vel.name].velocity == 100. * dt * dt
    assert pr2_world.state[omni_drive.y_vel.name].position == 0

    assert pr2_world.state[omni_drive.x.name].jerk == 0.
    assert pr2_world.state[omni_drive.x.name].acceleration == 0.
    assert pr2_world.state[omni_drive.x.name].velocity == 0.8951707486311977
    assert pr2_world.state[omni_drive.x.name].position == 0.08951707486311977

    assert pr2_world.state[omni_drive.y.name].jerk == 0.
    assert pr2_world.state[omni_drive.y.name].acceleration == 0.
    assert pr2_world.state[omni_drive.y.name].velocity == 1.094837581924854
    assert pr2_world.state[omni_drive.y.name].position == 0.1094837581924854

def test_pr2_view(pr2_world):
    pr2 = PR2.from_world(pr2_world)

    # Ensure there are no loose bodies
    pr2_world._notify_model_change()

    assert len(pr2.manipulators) == 2
    assert len(pr2.manipulator_chains) == 2
    assert len(pr2.sensors) == 1
    assert len(pr2.sensor_chains) == 1
    assert pr2.neck == list(pr2.sensor_chains)[0]
    assert pr2.torso.name.name == 'torso'
    assert len(pr2.torso.sensors) == 0
    assert list(pr2.sensor_chains)[0].sensors == pr2.sensors
    assert pr2.odom.name.name == 'odom_combined'



import inspect
import itertools
import logging
import os
import sys
import unittest

import sqlalchemy
from sqlalchemy import create_engine, select, text
from sqlalchemy.orm import Session

from semantic_world.adapters.urdf import URDFParser
from semantic_world.connections import RevoluteConnection
from semantic_world.geometry import Shape, Box, Scale, Color
from semantic_world.orm.model import WorldMapping, QuaternionMapping
from semantic_world.prefixed_name import PrefixedName
from semantic_world.spatial_types.spatial_types import TransformationMatrix
from semantic_world.world import Body
from semantic_world.orm.ormatic_interface import *
from ormatic.dao import to_dao


class ORMTest(unittest.TestCase):
    engine: sqlalchemy.engine
    session: Session

    urdf_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..", "resources", "urdf")
    table = os.path.join(urdf_dir, "table.urdf")

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.engine = create_engine('sqlite:///:memory:')
        cls.table_world = URDFParser(cls.table).parse()

    def setUp(self):
        super().setUp()
        self.session = Session(self.engine)
        Base.metadata.create_all(bind=self.session.bind)

    def tearDown(self):
        super().tearDown()
        Base.metadata.drop_all(self.session.bind)
        self.session.close()

    def test_table_world(self):
        world_dao: WorldMappingDAO = to_dao(self.table_world)

        self.session.add(world_dao)
        self.session.commit()

        bodies_from_db = self.session.scalars(select(BodyDAO)).all()
        self.assertEqual(len(bodies_from_db), len(self.table_world.bodies))

        connections_from_db = self.session.scalars(select(ConnectionDAO)).all()
        self.assertEqual(len(connections_from_db), len(self.table_world.connections))

        queried_world = self.session.scalar(select(WorldMappingDAO))
        reconstructed = queried_world.from_dao()
        

    def test_insert(self):
        origin = TransformationMatrix.from_xyz_rpy(1, 2, 3, 1, 2, 3)
        scale = Scale(1., 1., 1.)
        color = Color(0., 1., 1.)
        shape1 = Box(origin=origin, scale=scale, color=color)
        b1 = Body(
            name=PrefixedName("b1"),
            collision=[shape1]
        )

        dao: BodyDAO = to_dao(b1)

        self.session.add(dao)
        self.session.commit()
        queried_body = self.session.scalar(select(BodyDAO))
        reconstructed_body = queried_body.from_dao()
        self.assertIs(reconstructed_body, reconstructed_body.collision[0].origin.reference_frame)

        result = self.session.scalar(select(ShapeDAO))
        self.assertIsInstance(result, BoxDAO)
        box = result.from_dao()



from __future__ import annotations

import functools
from enum import IntEnum
from typing import overload, Union, Iterable, Tuple, Optional, Callable, List, Sequence, Dict, TypeVar, \
    TYPE_CHECKING
import numpy as np
import casadi as ca

from scipy import sparse as sp

if TYPE_CHECKING:
    from ..world_entity import Body

all_expressions = Union[Symbol_, Symbol, Expression, Point3, Vector3, RotationMatrix, TransformationMatrix, Quaternion]
all_expressions_float = Union[Symbol, Expression, Point3, Vector3, RotationMatrix, TransformationMatrix, float, Quaternion]
symbol_expr_float = Union[Symbol, Expression, float, int, IntEnum]
symbol_expr = Union[Symbol, Expression]
PreservedCasType = TypeVar('PreservedCasType', Point3, Vector3, TransformationMatrix, RotationMatrix, Quaternion, Expression)

pi: float

def _operation_type_error(arg1: object, operation: str, arg2: object) -> TypeError: ...

class ReferenceFrameMixin:
    reference_frame: Optional[Body]


class StackedCompiledFunction:
    compiled_f: CompiledFunction
    split_out_view: List[np.ndarray]
    symbol_parameters: List[Symbol]

    def __init__(self,
                 expressions: List[Expression],
                 parameters: Optional[Union[List[Symbol], List[List[Symbol]]]] = None,
                 additional_views: Optional[List[slice]] = None): ...

    def fast_call(self, *args: np.ndarray) -> np.ndarray: ...


class CompiledFunction:
    symbol_parameters: List[Symbol]
    compiled_casadi_function: ca.Function
    function_buffer: ca.FunctionBuffer
    function_evaluator: functools.partial
    out: Union[np.ndarray, sp.csc_matrix]
    sparse: bool

    def __init__(self,
                 expression: Symbol_,
                 parameters: Optional[Union[List[Symbol], List[List[Symbol]]]] = None,
                 sparse: bool = False): ...

    def __call__(self, **kwargs) -> np.ndarray: ...

    def fast_call(self, *args: np.ndarray) -> Union[np.ndarray, sp.csc_matrix]: ...


class Symbol_:
    s: Union[ca.SX, np.ndarray]
    np_data: Optional[np.ndarray]

    @property
    def shape(self) -> Tuple[int, int]: ...

    def __getitem__(self, item: Union[np.ndarray, Union[int, slice], Tuple[Union[int, slice], Union[int, slice]]]) -> Expression: ...

    def __setitem__(self, key: Union[Union[int, slice], Tuple[Union[int, slice], Union[int, slice]]], value: symbol_expr_float): ...

    def __len__(self) -> int: ...

    def free_symbols(self) -> List[ca.SX]: ...

    def is_constant(self) -> bool: ...

    def to_np(self) -> Union[float, np.ndarray]: ...

    def compile(self, parameters: Optional[Union[List[Symbol], List[List[Symbol]]]] = None, sparse: bool = False) \
            -> CompiledFunction: ...

    def __hash__(self) -> int: ...

    def pretty_str(self) -> str: ...


class Symbol(Symbol_):
    _registry: Dict[str, Symbol]
    name: str

    def __init__(self, name: str): ...

    @overload
    def __add__(self, other: Point3) -> Point3: ...
    @overload
    def __add__(self, other: Vector3) -> Vector3: ...
    @overload
    def __add__(self, other: Union[Symbol, Expression, float, Quaternion]) -> Expression: ...
    def __radd__(self, other: float) -> Expression: ...

    @overload
    def __sub__(self, other: Point3) -> Point3: ...
    @overload
    def __sub__(self, other: Vector3) -> Vector3: ...
    @overload
    def __sub__(self, other: RotationMatrix) -> RotationMatrix: ...
    @overload
    def __sub__(self, other: TransformationMatrix) -> TransformationMatrix: ...
    @overload
    def __sub__(self, other: Union[Symbol, Expression, float, Quaternion]) -> Expression: ...
    def __rsub__(self, other: float) -> Expression: ...

    @overload
    def __mul__(self, other: Point3) -> Point3: ...
    @overload
    def __mul__(self, other: Vector3) -> Vector3: ...
    @overload
    def __mul__(self, other: RotationMatrix) -> RotationMatrix: ...
    @overload
    def __mul__(self, other: TransformationMatrix) -> TransformationMatrix: ...
    @overload
    def __mul__(self, other: Union[Symbol, Expression, float, Quaternion]) -> Expression: ...
    def __rmul__(self, other: float) -> Expression: ...

    @overload
    def __truediv__(self, other: Point3) -> Point3: ...
    @overload
    def __truediv__(self, other: Vector3) -> Vector3: ...
    @overload
    def __truediv__(self, other: RotationMatrix) -> RotationMatrix: ...
    @overload
    def __truediv__(self, other: TransformationMatrix) -> TransformationMatrix: ...
    @overload
    def __truediv__(self, other: Union[Symbol, Expression, float, Quaternion]) -> Expression: ...
    def __rtruediv__(self, other: float) -> Expression: ...

    def __floordiv__(self, other: symbol_expr_float) -> Expression: ...
    def __rfloordiv__(self, other: symbol_expr_float) -> Expression: ...
    def __mod__(self, other: symbol_expr_float) -> Expression: ...
    def __rmod__(self, other: symbol_expr_float) -> Expression: ...
    def __divmod__(self, other: symbol_expr_float) -> Tuple[Expression, Expression]: ...
    def __rdivmod__(self, other: symbol_expr_float) -> Tuple[Expression, Expression]: ...

    def __lt__(self, other: all_expressions_float) -> Expression: ...
    def __le__(self, other: all_expressions_float) -> Expression: ...
    def __gt__(self, other: all_expressions_float) -> Expression: ...
    def __ge__(self, other: all_expressions_float) -> Expression: ...
    def __and__(self, other: all_expressions_float) -> Expression: ...
    def __or__(self, other: all_expressions_float) -> Expression: ...
    def __neg__(self) -> Expression: ...
    def __invert__(self) -> Expression: ...
    def __pow__(self, other: all_expressions_float) -> Expression: ...
    def __rpow__(self, other: all_expressions_float) -> Expression: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...


class Expression(Symbol_):
    def __init__(self, data: Optional[Union[Symbol,
                                            Expression,
                                            float,
                                            Vector3, Point3, TransformationMatrix, RotationMatrix, Quaternion,
                                            Iterable[symbol_expr_float],
                                            Iterable[Iterable[symbol_expr_float]],
                                            np.ndarray]] = None): ...

    @overload
    def __add__(self, other: Point3) -> Point3: ...
    @overload
    def __add__(self, other: Vector3) -> Vector3: ...
    @overload
    def __add__(self, other: Union[Symbol, Expression, float, TransformationMatrix, RotationMatrix, Quaternion]) -> Expression: ...
    def __radd__(self, other: float) -> Expression: ...

    @overload
    def __sub__(self, other: Point3) -> Point3: ...
    @overload
    def __sub__(self, other: Vector3) -> Vector3: ...
    @overload
    def __sub__(self, other: Union[Symbol, Expression, float, TransformationMatrix, RotationMatrix, Quaternion]) -> Expression: ...
    def __rsub__(self, other: float) -> Expression: ...

    @overload
    def __mul__(self, other: Point3) -> Point3: ...
    @overload
    def __mul__(self, other: Vector3) -> Vector3: ...
    @overload
    def __mul__(self, other: Union[Symbol, Expression, float, RotationMatrix, TransformationMatrix, Quaternion]) -> Expression: ...
    def __rmul__(self, other: float) -> Expression: ...

    @overload
    def __truediv__(self, other: Point3) -> Point3: ...
    @overload
    def __truediv__(self, other: Vector3) -> Vector3: ...
    @overload
    def __truediv__(self, other: Union[Symbol, Expression, float, RotationMatrix, TransformationMatrix, Quaternion]) -> Expression: ...
    def __rtruediv__(self, other: float) -> Expression: ...

    def __floordiv__(self, other: symbol_expr_float) -> Expression: ...
    def __rfloordiv__(self, other: symbol_expr_float) -> Expression: ...
    def __mod__(self, other: symbol_expr_float) -> Expression: ...
    def __rmod__(self, other: symbol_expr_float) -> Expression: ...
    def __divmod__(self, other: symbol_expr_float) -> Tuple[Expression, Expression]: ...
    def __rdivmod__(self, other: symbol_expr_float) -> Tuple[Expression, Expression]: ...

    def __eq__(self, other: symbol_expr_float) -> Expression: ...
    def __lt__(self, other: symbol_expr_float) -> Expression: ...
    def __le__(self, other: symbol_expr_float) -> Expression: ...
    def __gt__(self, other: symbol_expr_float) -> Expression: ...
    def __ge__(self, other: symbol_expr_float) -> Expression: ...
    def __and__(self, other: all_expressions_float) -> Expression: ...
    def __or__(self, other: all_expressions_float) -> Expression: ...
    def __neg__(self) -> Expression: ...
    def __invert__(self) -> Expression: ...
    def __pow__(self, other: symbol_expr_float) -> Expression: ...
    def __rpow__(self, other: symbol_expr_float) -> Expression: ...
    def __copy__(self) -> Expression: ...

    @property
    def T(self) -> Expression: ...

    def remove(self, rows: List[int], columns: List[int]): ...

    def split(self) -> List[Expression]: ...

    def dot(self, other: Expression) -> Expression: ...

    def reshape(self, new_shape: Tuple[int, int]) -> Expression: ...


class Point3(Symbol_, ReferenceFrameMixin):

    @property
    def x(self) -> Expression: ...
    @x.setter
    def x(self, value: symbol_expr_float): ...
    @property
    def y(self) -> Expression: ...
    @y.setter
    def y(self, value: symbol_expr_float): ...
    @property
    def z(self) -> Expression: ...
    @z.setter
    def z(self, value: symbol_expr_float): ...

    def __init__(self, x: symbol_expr_float = 0, y: symbol_expr_float = 0, z: symbol_expr_float = 0,
                 reference_frame: Optional[Body] = None): ...

    @classmethod
    def from_iterable(cls, data: Optional[Union[Expression, Point3, Vector3,
                                            ca.SX,
                                            np.ndarray,
                                            Iterable[symbol_expr_float]]] = None,
                 reference_frame: Optional[Body] = None) -> Point3: ...

    def norm(self) -> Expression: ...

    @overload
    def __add__(self, other: Union[Vector3, Symbol, Expression]) -> Point3: ...
    @overload
    def __add__(self, other: float) -> Point3: ...
    def __radd__(self, other: float) -> Point3: ...

    @overload
    def __sub__(self, other: Point3) -> Vector3: ...
    @overload
    def __sub__(self, other: Union[Symbol, Expression, Vector3]) -> Point3: ...
    @overload
    def __sub__(self, other: float) -> Point3: ...
    def __rsub__(self, other: float) -> Point3: ...

    @overload
    def __mul__(self, other: Symbol) -> Point3: ...
    @overload
    def __mul__(self, other: Expression) -> Point3: ...
    @overload
    def __mul__(self, other: float) -> Point3: ...
    def __rmul__(self, other: float) -> Point3: ...

    @overload
    def __matmul__(self, other: Point3) -> Expression: ...
    @overload
    def __matmul__(self, other: Vector3) -> Expression: ...

    @overload
    def __pow__(self, other: Symbol) -> Point3: ...
    @overload
    def __pow__(self, other: Expression) -> Point3: ...
    @overload
    def __pow__(self, other: float) -> Point3: ...
    def __rpow__(self, other: float) -> Point3: ...

    @overload
    def __truediv__(self, other: Symbol) -> Point3: ...
    @overload
    def __truediv__(self, other: Expression) -> Point3: ...
    @overload
    def __truediv__(self, other: float) -> Point3: ...
    def __rtruediv__(self, other: float) -> Point3: ...

    def __neg__(self) -> Point3: ...

    @overload
    def dot(self, other: Point3) -> Expression: ...
    @overload
    def dot(self, other: Vector3) -> Expression: ...


class Vector3(Symbol_, ReferenceFrameMixin):
    vis_frame: Optional[Body]

    @property
    def x(self) -> Expression: ...
    @x.setter
    def x(self, value: symbol_expr_float): ...
    @property
    def y(self) -> Expression: ...
    @y.setter
    def y(self, value: symbol_expr_float): ...
    @property
    def z(self) -> Expression: ...
    @z.setter
    def z(self, value: symbol_expr_float): ...

    def __init__(self, x: symbol_expr_float = 0, y: symbol_expr_float = 0, z: symbol_expr_float = 0,
                 reference_frame: Optional[Body] = None): ...

    @classmethod
    def from_iterable(cls, data: Optional[Union[Expression, Point3, Vector3,
                                            ca.SX,
                                            np.ndarray,
                                            Iterable[symbol_expr_float]]] = None,
                 reference_frame: Optional[Body] = None) -> Vector3: ...

    @classmethod
    def X(cls, reference_frame: Optional[Body] = None) -> Vector3: ...
    @classmethod
    def Y(cls, reference_frame: Optional[Body] = None) -> Vector3: ...
    @classmethod
    def Z(cls, reference_frame: Optional[Body] = None) -> Vector3: ...

    @classmethod
    def unit_vector(cls, x: symbol_expr_float = 0, y: symbol_expr_float = 0, z: symbol_expr_float = 0,
                    reference_frame: Optional[Body] = None) -> Vector3: ...

    def norm(self) -> Expression: ...

    def scale(self, a: symbol_expr_float, unsafe: bool = False): ...

    @overload
    def __add__(self, other: Point3) -> Point3: ...
    @overload
    def __add__(self, other: Vector3) -> Vector3: ...
    @overload
    def __add__(self, other: Symbol) -> Vector3: ...
    @overload
    def __add__(self, other: Expression) -> Vector3: ...
    @overload
    def __add__(self, other: float) -> Vector3: ...
    def __radd__(self, other: float) -> Vector3: ...

    @overload
    def __sub__(self, other: Vector3) -> Vector3: ...
    @overload
    def __sub__(self, other: Point3) -> Point3: ...
    @overload
    def __sub__(self, other: Symbol) -> Vector3: ...
    @overload
    def __sub__(self, other: Expression) -> Vector3: ...
    @overload
    def __sub__(self, other: float) -> Vector3: ...
    def __rsub__(self, other: float) -> Vector3: ...

    @overload
    def __mul__(self, other: Symbol) -> Vector3: ...
    @overload
    def __mul__(self, other: Expression) -> Vector3: ...
    @overload
    def __mul__(self, other: float) -> Vector3: ...
    def __rmul__(self, other: float) -> Vector3: ...

    @overload
    def __matmul__(self, other: Point3) -> Expression: ...
    @overload
    def __matmul__(self, other: Vector3) -> Expression: ...

    @overload
    def __truediv__(self, other: Symbol) -> Vector3: ...
    @overload
    def __truediv__(self, other: Expression) -> Vector3: ...
    @overload
    def __truediv__(self, other: float) -> Vector3: ...
    def __rtruediv__(self, other: float) -> Vector3: ...

    def __neg__(self) -> Vector3: ...

    @overload
    def __pow__(self, other: Symbol) -> Vector3: ...
    @overload
    def __pow__(self, other: Expression) -> Vector3: ...
    @overload
    def __pow__(self, other: float) -> Vector3: ...
    def __rpow__(self, other: float) -> Vector3: ...

    @overload
    def dot(self, other: Point3) -> Expression: ...
    @overload
    def dot(self, other: Vector3) -> Expression: ...

    def cross(self, other: Vector3) -> Vector3: ...


TrinaryFalse = 0
TrinaryUnknown = 0.5
TrinaryTrue = 1

BinaryTrue: Expression
BinaryFalse: Expression



class TransformationMatrix(Symbol_, ReferenceFrameMixin):
    child_frame: Optional[Body]

    def __init__(self, data: Optional[Union[TransformationMatrix,
                                            RotationMatrix,
                                            ca.SX,
                                            np.ndarray,
                                            Expression,
                                            Iterable[Iterable[symbol_expr_float]]]] = None,
                 reference_frame: Optional[Body] = None,
                 child_frame: Optional[Body] = None,
                 sanity_check: bool = True): ...

    @property
    def x(self) -> Expression: ...
    @x.setter
    def x(self, value: symbol_expr_float): ...
    @property
    def y(self) -> Expression: ...
    @y.setter
    def y(self, value: symbol_expr_float): ...
    @property
    def z(self) -> Expression: ...
    @z.setter
    def z(self, value: symbol_expr_float): ...

    @overload
    def dot(self, other: Point3) -> Point3: ...
    @overload
    def dot(self, other: Vector3) -> Vector3: ...
    @overload
    def dot(self, other: RotationMatrix) -> RotationMatrix: ...
    @overload
    def dot(self, other: TransformationMatrix) -> TransformationMatrix: ...

    @overload
    def __matmul__(self, other: Point3) -> Point3: ...
    @overload
    def __matmul__(self, other: Vector3) -> Vector3: ...
    @overload
    def __matmul__(self, other: RotationMatrix) -> RotationMatrix: ...
    @overload
    def __matmul__(self, other: TransformationMatrix) -> TransformationMatrix: ...

    def to_rotation(self) -> RotationMatrix: ...
    def to_quaternion(self) -> Quaternion: ...
    def to_position(self) -> Point3: ...
    def to_translation(self) -> TransformationMatrix: ...

    @classmethod
    def from_xyz_rpy(cls,
                     x: Optional[symbol_expr_float] = 0,
                     y: Optional[symbol_expr_float] = 0,
                     z: Optional[symbol_expr_float] = 0,
                     roll: Optional[symbol_expr_float] = 0,
                     pitch: Optional[symbol_expr_float] = 0,
                     yaw: Optional[symbol_expr_float] = 0,
                     reference_frame: Optional[Body] = None,
                     child_frame: Optional[Body] = None) -> TransformationMatrix: ...
    @classmethod
    def from_xyz_quat(cls,
                      pos_x: symbol_expr_float = 0, pos_y: symbol_expr_float = 0, pos_z: symbol_expr_float =00,
                      quat_w: symbol_expr_float = 0, quat_x: symbol_expr_float = 0,
                      quat_y: symbol_expr_float = 0, quat_z: symbol_expr_float = 1,
                      reference_frame: Optional[Body] = None, child_frame: Optional[Body] = None) \
            -> TransformationMatrix: ...
    @classmethod
    def from_point_rotation_matrix(cls,
                                   point: Optional[Point3] = None,
                                   rotation_matrix: Optional[RotationMatrix] = None,
                                   reference_frame: Optional[Body] = None,
                                   child_frame: Optional[Body] = None) -> TransformationMatrix: ...

    def inverse(self) -> TransformationMatrix: ...


class RotationMatrix(Symbol_, ReferenceFrameMixin):
    child_frame: Optional[Body]

    def __init__(self, data: Optional[Union[TransformationMatrix,
                                            RotationMatrix,
                                            Expression,
                                            Quaternion,
                                            ca.SX,
                                            np.ndarray,
                                            Iterable[Iterable[symbol_expr_float]]]] = None,
                 reference_frame: Optional[Body] = None,
                 child_frame: Optional[Body] = None,
                 sanity_check: bool = True): ...
    @classmethod
    def __quaternion_to_rotation_matrix(cls, q: Quaternion) -> RotationMatrix: ...

    @classmethod
    def from_axis_angle(cls, axis: Vector3, angle: symbol_expr_float, reference_frame: Optional[Body] = None) \
            -> RotationMatrix: ...
    @classmethod
    def from_quaternion(cls, q: Quaternion) -> RotationMatrix: ...
    @classmethod
    @overload
    def from_vectors(cls, x: Vector3, y: Vector3, reference_frame: Optional[Body] = None) -> RotationMatrix: ...
    @classmethod
    @overload
    def from_vectors(cls, x: Vector3, z: Vector3, reference_frame: Optional[Body] = None) -> RotationMatrix: ...
    @classmethod
    @overload
    def from_vectors(cls, y: Vector3, z: Vector3, reference_frame: Optional[Body] = None) -> RotationMatrix: ...
    @classmethod
    @overload
    def from_vectors(cls, x: Vector3, y: Vector3, z: Vector3,
                     reference_frame: Optional[Body] = None) -> RotationMatrix: ...

    @classmethod
    def from_rpy(cls,
                 roll: Optional[symbol_expr_float] = None,
                 pitch: Optional[symbol_expr_float] = None,
                 yaw: Optional[symbol_expr_float] = None,
                 reference_frame: Optional[Body] = None) -> RotationMatrix: ...

    def x_vector(self) -> Vector3: ...
    def y_vector(self) -> Vector3: ...
    def z_vector(self) -> Vector3: ...

    def to_axis_angle(self) -> Tuple[Vector3, Expression]: ...
    def to_angle(self, hint: Optional[Callable] = None) -> Expression: ...
    def to_rpy(self) -> Tuple[Expression, Expression, Expression]: ...
    def to_quaternion(self) -> Quaternion: ...

    @overload
    def dot(self, other: Vector3) -> Vector3: ...
    @overload
    def dot(self, other: RotationMatrix) -> RotationMatrix: ...
    @overload
    def dot(self, other: TransformationMatrix) -> TransformationMatrix: ...

    @overload
    def __matmul__(self, other: Vector3) -> Vector3: ...
    @overload
    def __matmul__(self, other: RotationMatrix) -> RotationMatrix: ...
    @overload
    def __matmul__(self, other: TransformationMatrix) -> TransformationMatrix: ...

    def normalize(self): ...
    def inverse(self) -> RotationMatrix: ...
    @property
    def T(self) -> RotationMatrix: ...


class Quaternion(Symbol_, ReferenceFrameMixin):
    @property
    def x(self) -> Expression: ...
    @x.setter
    def x(self, value: symbol_expr_float): ...
    @property
    def y(self) -> Expression: ...
    @y.setter
    def y(self, value: symbol_expr_float): ...
    @property
    def z(self) -> Expression: ...
    @z.setter
    def z(self, value: symbol_expr_float): ...
    @property
    def w(self) -> Expression: ...
    @w.setter
    def w(self, value: symbol_expr_float): ...

    def __init__(self, x: symbol_expr_float = 0.0, y: symbol_expr_float = 0.0,
                 z: symbol_expr_float = 0.0, w: symbol_expr_float = 1.0,
                 reference_frame: Optional[Body] = None): ...

    @classmethod
    def from_iterable(cls, data: Optional[Union[Expression,
                                            Quaternion,
                                            ca.SX,
                                            Tuple[symbol_expr_float,
                                                  symbol_expr_float,
                                                  symbol_expr_float,
                                                  symbol_expr_float]]] = None,
                 reference_frame: Optional[Body] = None) -> Quaternion: ...

    @classmethod
    def from_axis_angle(cls, axis: Vector3, angle: symbol_expr_float, reference_frame: Optional[Body] = None) \
            -> Quaternion: ...
    @classmethod
    def from_rpy(cls, roll: symbol_expr_float, pitch: symbol_expr_float, yaw: symbol_expr_float,
                 reference_frame: Optional[Body] = None) -> Quaternion: ...
    @classmethod
    def from_rotation_matrix(cls, r: Union[RotationMatrix, TransformationMatrix]) -> Quaternion: ...

    def conjugate(self) -> Quaternion: ...
    def multiply(self, q: Quaternion) -> Quaternion: ...
    def diff(self, q: Quaternion) -> Quaternion: ...
    def norm(self) -> Expression: ...
    def normalize(self): ...
    def to_axis_angle(self) -> Tuple[Vector3, Expression]: ...
    def to_rotation_matrix(self) -> RotationMatrix: ...
    def to_rpy(self) -> Tuple[Expression, Expression, Expression]: ...
    def dot(self, other: Quaternion) -> Expression: ...

    def __neg__(self) -> Quaternion: ...


def norm(v: Union[Vector3, Point3, Expression, Quaternion]) -> Expression: ...

@overload
def save_division(nominator: Vector3, denominator: symbol_expr_float, if_nan: Optional[Vector3] = None) -> Vector3: ...
@overload
def save_division(nominator: Point3, denominator: symbol_expr_float, if_nan: Optional[Point3] = None) -> Point3: ...
@overload
def save_division(nominator: symbol_expr_float, denominator: symbol_expr_float,
                  if_nan: Optional[symbol_expr_float] = None) -> Expression: ...

def diag(args: Union[List[symbol_expr_float], Expression]) -> Expression: ...

def hessian(expressions: Union[symbol_expr, List[symbol_expr]], symbols: Iterable[Symbol]) -> Expression: ...

def jacobian(expressions: Union[symbol_expr, List[symbol_expr]], symbols: Iterable[Symbol]) -> Expression: ...

def jacobian_dot(expressions: Expression, symbols: List[symbol_expr], symbols_dot: List[symbol_expr]) -> Expression: ...

def jacobian_ddot(expressions: Expression, symbols: List[symbol_expr], symbols_dot: List[symbol_expr],
                  symbols_ddot: List[symbol_expr]) -> Expression: ...

def equivalent(expression1: symbol_expr, expression2: symbol_expr) -> bool: ...

def var(variables_names: str) -> List[Symbol]: ...

def free_symbols(expression: all_expressions) -> List[ca.SX]: ...

def create_symbols(names: Union[List[str], int]) -> List[Symbol]: ...

def compile_and_execute(f: Callable[..., all_expressions],
                        params: Union[List[Union[float, np.ndarray]], np.ndarray]) \
        -> Union[float, np.ndarray]: ...

def zeros(rows: int, columns: int) -> Expression: ...

def ones(x: int, y: int) -> Expression: ...

def tri(dimension: int) -> Expression: ...

def eye(size: int) -> Expression: ...

@overload
def abs(x: Vector3) -> Vector3: ...
@overload
def abs(x: Point3) -> Point3: ...
@overload
def abs(x: symbol_expr_float) -> Expression: ...

@overload
def max(x: Point3) -> Expression: ...
@overload
def max(x: Vector3) -> Expression: ...
@overload
def max(x: symbol_expr_float, y: Optional[symbol_expr_float] = None) -> Expression: ...

@overload
def min(x: Point3) -> Expression: ...
@overload
def min(x: Vector3) -> Expression: ...
@overload
def min(x: symbol_expr_float, y: Optional[symbol_expr_float] = None) -> Expression: ...

def limit(x: symbol_expr_float,
          lower_limit: symbol_expr_float,
          upper_limit: symbol_expr_float) -> Expression: ...

def equal(x: symbol_expr_float, y: symbol_expr_float) -> Expression: ...

def not_equal(x: symbol_expr_float, y: symbol_expr_float) -> Expression: ...

def less_equal(x: symbol_expr_float, y: symbol_expr_float) -> Expression: ...
def greater_equal(x: symbol_expr_float, y: symbol_expr_float) -> Expression: ...
def less(x: symbol_expr_float, y: symbol_expr_float) -> Expression: ...
def greater(x: symbol_expr_float, y: symbol_expr_float) -> Expression: ...

def logic_and(*args: symbol_expr_float) -> symbol_expr_float: ...

def logic_and3(*args: symbol_expr_float) -> symbol_expr_float: ...

def logic_or(*args: symbol_expr_float, simplify: bool = True) -> symbol_expr_float: ...

def logic_or3(*args: symbol_expr_float) -> symbol_expr_float: ...

def logic_any(args: Expression) -> symbol_expr_float: ...

def logic_all(args: Expression) -> symbol_expr_float: ...

def logic_not(expr: symbol_expr_float) -> Expression: ...

def logic_not3(expr: symbol_expr_float) -> Expression: ...

@overload
def if_else(condition: symbol_expr_float, if_result: Vector3, else_result: Vector3) -> Vector3: ...
@overload
def if_else(condition: symbol_expr_float, if_result: Point3, else_result: Point3) -> Point3: ...
@overload
def if_else(condition: symbol_expr_float, if_result: symbol_expr_float,
            else_result: symbol_expr_float) -> Expression: ...

@overload
def if_greater(a: symbol_expr_float, b: symbol_expr_float, if_result: Vector3,
               else_result: Vector3) -> Vector3: ...
@overload
def if_greater(a: symbol_expr_float, b: symbol_expr_float, if_result: Point3,
               else_result: Point3) -> Point3: ...
@overload
def if_greater(a: symbol_expr_float, b: symbol_expr_float, if_result: symbol_expr_float,
               else_result: symbol_expr_float) -> Expression: ...

@overload
def if_less(a: symbol_expr_float, b: symbol_expr_float, if_result: Quaternion,
            else_result: Quaternion) -> Quaternion: ...
@overload
def if_less(a: symbol_expr_float, b: symbol_expr_float, if_result: Vector3,
            else_result: Vector3) -> Vector3: ...
@overload
def if_less(a: symbol_expr_float, b: symbol_expr_float, if_result: Point3,
            else_result: Point3) -> Point3: ...
@overload
def if_less(a: symbol_expr_float, b: symbol_expr_float, if_result: symbol_expr_float,
            else_result: symbol_expr_float) -> Expression: ...

@overload
def if_greater_zero(condition: symbol_expr_float,
                    if_result: Vector3,
                    else_result: Vector3) -> Vector3: ...
@overload
def if_greater_zero(condition: symbol_expr_float,
                    if_result: Point3,
                    else_result: Point3) -> Point3: ...
@overload
def if_greater_zero(condition: symbol_expr_float,
                    if_result: Quaternion,
                    else_result: Quaternion) -> Quaternion: ...
@overload
def if_greater_zero(condition: symbol_expr_float,
                    if_result: symbol_expr_float,
                    else_result: symbol_expr_float) -> Expression: ...

@overload
def if_greater_eq_zero(condition: symbol_expr_float, if_result: Vector3,
                       else_result: Vector3) -> Vector3: ...
@overload
def if_greater_eq_zero(condition: symbol_expr_float, if_result: Point3,
                       else_result: Point3) -> Point3: ...
@overload
def if_greater_eq_zero(condition: symbol_expr_float, if_result: symbol_expr_float,
                       else_result: symbol_expr_float) -> Expression: ...

@overload
def if_greater_eq(a: symbol_expr_float, b: symbol_expr_float, if_result: Vector3,
                  else_result: Vector3) -> Vector3: ...
@overload
def if_greater_eq(a: symbol_expr_float, b: symbol_expr_float, if_result: Point3,
                  else_result: Point3) -> Point3: ...
@overload
def if_greater_eq(a: symbol_expr_float, b: symbol_expr_float, if_result: symbol_expr_float,
                  else_result: symbol_expr_float) -> Expression: ...

@overload
def if_less_eq(a: symbol_expr_float, b: symbol_expr_float, if_result: Vector3,
               else_result: Vector3) -> Vector3: ...
@overload
def if_less_eq(a: symbol_expr_float, b: symbol_expr_float, if_result: Point3,
               else_result: Point3) -> Point3: ...
@overload
def if_less_eq(a: symbol_expr_float, b: symbol_expr_float, if_result: symbol_expr_float,
               else_result: symbol_expr_float) -> Expression: ...

@overload
def if_eq_zero(condition: symbol_expr_float,
               if_result: Vector3,
               else_result: Vector3) -> Vector3: ...
@overload
def if_eq_zero(condition: symbol_expr_float,
               if_result: Point3,
               else_result: Point3) -> Point3: ...
@overload
def if_eq_zero(condition: symbol_expr_float,
               if_result: symbol_expr_float,
               else_result: symbol_expr_float) -> Expression: ...

@overload
def if_eq(a: symbol_expr_float,
          b: symbol_expr_float,
          if_result: Vector3,
          else_result: Vector3) -> Vector3: ...
@overload
def if_eq(a: symbol_expr_float,
          b: symbol_expr_float,
          if_result: Point3,
          else_result: Point3) -> Point3: ...
@overload
def if_eq(a: symbol_expr_float,
          b: symbol_expr_float,
          if_result: symbol_expr_float,
          else_result: symbol_expr_float) -> Expression: ...

def if_eq_cases(a: symbol_expr_float,
                b_result_cases: Iterable[Tuple[symbol_expr_float, symbol_expr_float]],
                else_result: symbol_expr_float) -> Expression: ...

def if_eq_cases_grouped(a: symbol_expr_float,
                        b_result_cases: Iterable[Tuple[symbol_expr_float, symbol_expr_float]],
                        else_result: symbol_expr_float) -> Expression: ...

def if_cases(cases: Sequence[Tuple[symbol_expr_float, symbol_expr_float]],
             else_result: symbol_expr_float) -> Expression: ...

def if_less_eq_cases(a: symbol_expr_float,
                     b_result_cases: Sequence[Tuple[symbol_expr_float, symbol_expr_float]],
                     else_result: symbol_expr_float) -> Expression: ...

def cross(u: Union[Vector3, Expression], v: Union[Vector3, Expression]) -> Vector3: ...

def _to_sx(thing: object) -> ca.SX: ...

@overload
def scale(v: Vector3, a: symbol_expr_float) -> Vector3: ...
@overload
def scale(v: Point3, a: symbol_expr_float) -> Point3: ...
@overload
def scale(v: Expression, a: symbol_expr_float) -> Expression: ...

@overload
def dot(e1: TransformationMatrix, e2: Point3) -> Point3: ...
@overload
def dot(e1: TransformationMatrix, e2: Vector3) -> Vector3: ...
@overload
def dot(e1: RotationMatrix, e2: Point3) -> Point3: ...
@overload
def dot(e1: RotationMatrix, e2: Vector3) -> Vector3: ...
@overload
def dot(e1: RotationMatrix, e2: RotationMatrix) -> RotationMatrix: ...
@overload
def dot(e1: RotationMatrix, e2: TransformationMatrix) -> TransformationMatrix: ...
@overload
def dot(e1: TransformationMatrix, e2: RotationMatrix) -> TransformationMatrix: ...
@overload
def dot(e1: TransformationMatrix, e2: TransformationMatrix) -> TransformationMatrix: ...
@overload
def dot(e1: Quaternion, e2: Quaternion) -> Expression: ...
@overload
def dot(e1: Union[Vector3, Point3], e2: Union[Vector3, Point3]) -> Expression: ...
@overload
def dot(e1: Expression, e2: Expression) -> Expression: ...

def kron(m1: Expression, m2: Expression) -> Expression: ...

def trace(matrix: Union[Expression, RotationMatrix, TransformationMatrix]) -> Expression: ...

# def rotation_distance(a_R_b: Expression, a_R_c: Expression) -> Expression: ...

@overload
def vstack(list_of_matrices: List[Union[Point3, Vector3, Quaternion]]) -> Expression: ...
@overload
def vstack(list_of_matrices: List[TransformationMatrix]) -> Expression: ...
@overload
def vstack(list_of_matrices: List[Expression]) -> Expression: ...

@overload
def hstack(list_of_matrices: List[TransformationMatrix]) -> Expression: ...
@overload
def hstack(list_of_matrices: List[Expression]) -> Expression: ...

@overload
def diag_stack(list_of_matrices: List[TransformationMatrix]) -> Expression: ...
@overload
def diag_stack(list_of_matrices: List[Expression]) -> Expression: ...

def normalize_axis_angle(axis: Vector3, angle: symbol_expr_float) -> Tuple[Vector3, Expression]: ...

def axis_angle_from_rpy(roll: symbol_expr_float, pitch: symbol_expr_float, yaw: symbol_expr_float) \
        -> Tuple[Vector3, Expression]: ...

def cosine_distance(v0: symbol_expr_float, v1: symbol_expr_float) -> Expression: ...

@overload
def euclidean_distance(v1: symbol_expr_float, v2: symbol_expr_float) -> Expression: ...
@overload
def euclidean_distance(v1: Point3, v2: Point3) -> Expression: ...

def fmod(a: symbol_expr_float, b: symbol_expr_float) -> Expression: ...

def normalize_angle_positive(angle: symbol_expr_float) -> Expression: ...

def normalize_angle(angle: symbol_expr_float) -> Expression: ...

def shortest_angular_distance(from_angle: symbol_expr_float, to_angle: symbol_expr_float) -> Expression: ...

def quaternion_slerp(q1: Quaternion, q2: Quaternion, t: symbol_expr_float) -> Quaternion: ...

@overload
def slerp(v1: Vector3, v2: Vector3, t: symbol_expr_float) -> Vector3: ...
@overload
def slerp(v1: Expression, v2: Expression, t: symbol_expr_float) -> Expression: ...

def save_acos(angle: symbol_expr_float) -> Expression: ...

def entrywise_product(matrix1: Expression, matrix2: Expression) -> Expression: ...

def floor(x: symbol_expr_float) -> Expression: ...

def ceil(x: symbol_expr_float) -> Expression: ...

def round_up(x: symbol_expr_float, decimal_places: symbol_expr_float) -> Expression: ...

def round_down(x: symbol_expr_float, decimal_places: symbol_expr_float) -> Expression: ...

def sum(matrix: Expression) -> Expression: ...

def sum_row(matrix: Expression) -> Expression: ...

def sum_column(matrix: Expression) -> Expression: ...

def distance_point_to_line_segment(frame_P_current: Point3, frame_P_line_start: Point3, frame_P_line_end: Point3) \
        -> Tuple[Expression, Point3]: ...

def distance_point_to_line(frame_P_point: Point3, frame_P_line_point: Point3, frame_V_line_direction: Vector3) \
    -> Expression: ...

def distance_point_to_plane(frame_P_current: Point3, frame_V_v1: Vector3,
                                          frame_V_v2: Vector3) -> \
        Tuple[Expression, Point3]: ...

def distance_point_to_plane_signed(frame_P_current: Point3, frame_V_v1: Vector3,
                                          frame_V_v2: Vector3) -> \
        Tuple[Expression, Point3]: ...

def project_to_cone(frame_V_current: Vector3, frame_V_cone_axis: Vector3, cone_theta: Union[Symbol, float, Expression]) -> Vector3: ...

def project_to_plane(frame_V_plane_vector1: Vector3, frame_V_plane_vector2: Vector3, frame_P_point: Point3) -> Point3: ...

def angle_between_vector(v1: Vector3, v2: Vector3) -> Expression: ...

def velocity_limit_from_position_limit(acceleration_limit: Union[Symbol, float],
                                       position_limit: Union[Symbol, float],
                                       current_position: Union[Symbol, float],
                                       step_size: Union[Symbol, float],
                                       eps: float = 1e-5) -> Expression: ...

def rotational_error(r1: RotationMatrix, r2: RotationMatrix) -> Expression: ...

def to_str(expression: all_expressions) -> List[List[str]]: ...

def total_derivative(expr: Union[Symbol, Expression],
                     symbols: Iterable[Symbol],
                     symbols_dot: Iterable[Symbol]) \
        -> Expression: ...

def total_derivative2(expr: Union[Symbol, Expression],
                      symbols: Iterable[Symbol],
                      symbols_dot: Iterable[Symbol],
                      symbols_ddot: Iterable[Symbol]) -> Expression: ...

def quaternion_multiply(q1: Quaternion, q2: Quaternion) -> Quaternion: ...

def quaternion_conjugate(q: Quaternion) -> Quaternion: ...

def quaternion_diff(q1: Quaternion, q2: Quaternion) -> Quaternion: ...

def sign(x: symbol_expr_float) -> Expression: ...

def cos(x: symbol_expr_float) -> Expression: ...

def sin(x: symbol_expr_float) -> Expression: ...

def exp(x: symbol_expr_float) -> Expression: ...

def log(x: symbol_expr_float) -> Expression: ...

def tan(x: symbol_expr_float) -> Expression: ...

def sinh(x: symbol_expr_float) -> Expression: ...

def cosh(x: symbol_expr_float) -> Expression: ...

def sqrt(x: symbol_expr_float) -> Expression: ...

def acos(x: symbol_expr_float) -> Expression: ...

def atan2(x: symbol_expr_float, y: symbol_expr_float) -> Expression: ...

def solve_for(expression: Expression, target_value: float, start_value: float = 0.0001, max_tries: int = 10000,
              eps: float = 1e-10, max_step: float = 1) -> float: ...

def one_step_change(current_acceleration: symbol_expr_float, jerk_limit: symbol_expr_float, dt: float) \
        -> Expression: ...

def gauss(n: symbol_expr_float) -> Expression: ...


def r_gauss(integral: symbol_expr_float) -> Expression: ...


def substitute(expression: Union[Symbol, Expression], old_symbols: List[Symbol], new_symbols: List[Union[Symbol, Expression]]) -> Expression:
    ...

def matrix_inverse(a: Expression) -> Expression: ...

def gradient(ex: Expression, arg:Expression) -> Expression: ...

def is_true_symbol(expr: Expression) -> bool: ...

def is_true3(expr: Union[Symbol, Expression]) -> Expression: ...
def is_true3_symbol(expr: Expression) -> Expression: ...

def is_false3(expr: Union[Symbol, Expression]) -> Expression: ...
def is_false3_symbol(expr: Expression) -> Expression: ...

def is_unknown3(expr: Union[Symbol, Expression]) -> Expression: ...
def is_unknown3_symbol(expr: Expression) -> Expression: ...

def is_false_symbol(expr: Expression) -> bool: ...

def is_constant(expr: Expression) -> bool: ...

def det(expr: Expression) -> Expression: ...

def distance_projected_on_vector(point1: Point3, point2: Point3, vector: Vector3) -> Expression: ...

def distance_vector_projected_on_plane(point1: Point3, point2:Point3, normal_vector: Vector3) -> Vector3: ...

def replace_with_three_logic(expr: Expression) -> Expression: ...

def is_inf(expr: Expression) -> bool: ...

SpatialType = TypeVar('SpatialType', Point3, TransformationMatrix, Vector3, Quaternion, RotationMatrix)


from typing import Tuple
from unittest.mock import Mock

import numpy as np
import pytest

import semantic_world.spatial_types.spatial_types as cas
from semantic_world.spatial_types.symbol_manager import SymbolManager, symbol_manager


class TestSymbolManager:

    def test_singleton_behavior(self):
        """Test that SymbolManager follows singleton pattern"""
        manager1 = SymbolManager()
        manager2 = SymbolManager()
        assert manager1 is manager2
        assert manager1 is symbol_manager

    def test_init(self):
        """Test proper initialization of SymbolManager"""
        manager = SymbolManager()
        assert hasattr(manager, 'symbol_to_provider')
        assert isinstance(manager.symbol_to_provider, dict)

    def test_register_symbol_provider_with_callable(self):
        """Test registering symbols with callable providers"""
        manager = SymbolManager()

        def provider():
            return 42.0

        symbol = manager.register_symbol_provider('test_symbol', provider)

        assert isinstance(symbol, cas.Symbol)
        assert symbol in manager.symbol_to_provider
        assert manager.symbol_to_provider[symbol] == provider

    def test_register_symbol_provider_with_float(self):
        """Test registering symbols with float values"""
        manager = SymbolManager()

        symbol = manager.register_symbol_provider('test_float', 3.14)

        assert isinstance(symbol, cas.Symbol)
        assert symbol in manager.symbol_to_provider
        assert manager.symbol_to_provider[symbol] == 3.14

    def test_register_point3(self):
        """Test registering 3D points"""
        manager = SymbolManager()

        def point_provider() -> Tuple[float, float, float]:
            return (1.0, 2.0, 3.0)

        point = manager.register_point3('test_point', point_provider)

        assert isinstance(point, cas.Point3)
        # Check that x, y, z symbols were registered
        x_symbol = None
        y_symbol = None
        z_symbol = None

        for symbol in manager.symbol_to_provider.keys():
            if hasattr(symbol, 'name'):
                if symbol.name == 'test_point.x':
                    x_symbol = symbol
                elif symbol.name == 'test_point.y':
                    y_symbol = symbol
                elif symbol.name == 'test_point.z':
                    z_symbol = symbol

        assert x_symbol is not None
        assert y_symbol is not None
        assert z_symbol is not None

        # Test that providers return correct values
        assert manager.symbol_to_provider[x_symbol]() == 1.0
        assert manager.symbol_to_provider[y_symbol]() == 2.0
        assert manager.symbol_to_provider[z_symbol]() == 3.0

    def test_register_vector3(self):
        """Test registering 3D vectors"""
        manager = SymbolManager()

        def vector_provider() -> Tuple[float, float, float]:
            return (4.0, 5.0, 6.0)

        vector = manager.register_vector3('test_vector', vector_provider)

        assert isinstance(vector, cas.Vector3)
        # Similar checks as for point3
        symbols_found = 0
        for symbol in manager.symbol_to_provider.keys():
            if hasattr(symbol, 'name'):
                name = symbol.name
                if name in ['test_vector.x', 'test_vector.y', 'test_vector.z']:
                    symbols_found += 1

        assert symbols_found == 3

    def test_register_quaternion(self):
        """Test registering quaternions"""
        manager = SymbolManager()

        def quat_provider() -> Tuple[float, float, float, float]:
            return (0.0, 0.0, 0.0, 1.0)

        quaternion = manager.register_quaternion('test_quat', quat_provider)

        assert isinstance(quaternion, cas.Quaternion)
        # Check that x, y, z, w symbols were registered
        symbols_found = 0
        for symbol in manager.symbol_to_provider.keys():
            if hasattr(symbol, 'name'):
                name = symbol.name
                if name in ['test_quat.x', 'test_quat.y', 'test_quat.z', 'test_quat.w']:
                    symbols_found += 1

        assert symbols_found == 4

    def test_register_transformation_matrix(self):
        """Test registering transformation matrices"""
        manager = SymbolManager()

        def matrix_provider():
            return (
                (1.0, 0.0, 0.0, 0.1),
                (0.0, 1.0, 0.0, 0.2),
                (0.0, 0.0, 1.0, 0.3)
            )

        trans_matrix = manager.register_transformation_matrix('test_matrix', matrix_provider)

        assert isinstance(trans_matrix, cas.TransformationMatrix)
        # Check that 12 symbols were registered (3x4 matrix)
        matrix_symbols = 0
        for symbol in manager.symbol_to_provider.keys():
            if hasattr(symbol, 'name'):
                name = symbol.name
                if 'test_matrix[' in name:
                    matrix_symbols += 1

        assert matrix_symbols == 12

    def test_resolve_symbols_single_list(self):
        """Test resolving a single list of symbols"""
        manager = SymbolManager()

        symbol1 = manager.register_symbol_provider('test1', lambda: 1.0)
        symbol2 = manager.register_symbol_provider('test2', lambda: 2.0)

        result = manager.resolve_symbols([symbol1, symbol2])

        assert isinstance(result, np.ndarray)
        np.testing.assert_array_equal(result, np.array([1.0, 2.0]))

    def test_resolve_symbols_nested_list(self):
        """Test resolving nested lists of symbols"""
        manager = SymbolManager()

        symbol1 = manager.register_symbol_provider('test1', lambda: 1.0)
        symbol2 = manager.register_symbol_provider('test2', lambda: 2.0)
        symbol3 = manager.register_symbol_provider('test3', lambda: 3.0)

        result = manager.resolve_symbols([[symbol1, symbol2], [symbol3]])

        assert isinstance(result, list)
        assert len(result) == 2
        np.testing.assert_array_equal(result[0], np.array([1.0, 2.0]))
        np.testing.assert_array_equal(result[1], np.array([3.0]))

    def test_resolve_symbols_empty_list(self):
        """Test resolving empty symbol lists"""
        manager = SymbolManager()

        result = manager.resolve_symbols([])

        assert isinstance(result, np.ndarray)
        assert len(result) == 0

    def test_resolve_symbols_missing_symbol(self):
        """Test error handling for unregistered symbols"""
        manager = SymbolManager()

        # Create a symbol that's not registered
        unregistered_symbol = cas.Symbol('unregistered')

        with pytest.raises(KeyError, match='Cannot resolve'):
            manager.resolve_symbols([unregistered_symbol])


    def test_evaluate_expr_with_number(self):
        """Test evaluating numeric expressions"""
        manager = SymbolManager()

        result_int = manager.evaluate_expr(42)
        result_float = manager.evaluate_expr(3.14)

        assert result_int == 42
        assert result_float == 3.14

    def test_evaluate_expr_with_symbols(self):
        """Test evaluating symbolic expressions"""
        manager = SymbolManager()

        # Create a mock expression
        mock_expr = Mock()
        mock_compile = Mock()
        mock_compile.symbol_parameters = []
        mock_compile.fast_call.return_value = [5.0]
        mock_expr.compile.return_value = mock_compile
        mock_expr.to_np.return_value = 5.0

        result = manager.evaluate_expr(mock_expr)

        assert result == 5.0

    def test_evaluate_expr_no_params(self):
        """Test evaluating expressions without parameters"""
        manager = SymbolManager()

        # Create a mock expression with no parameters
        mock_expr = Mock()
        mock_compile = Mock()
        mock_compile.symbol_parameters = []
        mock_expr.compile.return_value = mock_compile
        mock_expr.to_np.return_value = 10.0

        result = manager.evaluate_expr(mock_expr)

        assert result == 10.0
        mock_expr.to_np.assert_called_once()

    def test_evaluate_expr_multiple_results(self):
        """Test evaluating expressions that return multiple values"""
        manager = SymbolManager()

        # Create a mock expression that returns multiple values
        expr = cas.Expression([1.0, 2.0, 3.0])

        result = manager.evaluate_expr(expr)

        assert np.allclose(result, [1.0, 2.0, 3.0])

    def test_provider_exception_handling(self):
        """Test that provider exceptions are properly handled"""
        manager = SymbolManager()

        def failing_provider():
            raise ValueError("Provider failed")

        symbol = manager.register_symbol_provider('failing', failing_provider)

        with pytest.raises(KeyError, match='Cannot resolve'):
            manager.resolve_symbols([symbol])

    def test_symbol_name_consistency(self):
        """Test that symbol names are consistent with registration names"""
        manager = SymbolManager()

        symbol = manager.register_symbol_provider('test_name', lambda: 1.0)

        # Note: This test assumes the symbol has a name() method
        # The actual implementation may vary based on the cas.Symbol class
        if hasattr(symbol, 'name'):
            assert symbol.name == 'test_name'


from __future__ import annotations

from dataclasses import dataclass, field
from functools import lru_cache
from typing import Optional, Union

from .spatial_types import spatial_types as cas
from .spatial_types.derivatives import Derivatives, DerivativeMap
from .spatial_types.symbol_manager import symbol_manager
from .world_entity import WorldEntity


@dataclass
class DegreeOfFreedom(WorldEntity):
    """
    A class representing a degree of freedom in a world model with associated derivatives and limits.
    
    This class manages a variable that can freely change within specified limits, tracking its position,
    velocity, acceleration, and jerk. It maintains symbolic representations for each derivative order
    and provides methods to get and set limits for these derivatives.
    """

    lower_limits: DerivativeMap[float] = field(default_factory=DerivativeMap)
    upper_limits: DerivativeMap[float] = field(default_factory=DerivativeMap)
    """
    Lower and upper bounds for each derivative
    """

    symbols: DerivativeMap[cas.Symbol] = field(default_factory=DerivativeMap, init=False)
    """
    Symbolic representations for each derivative
    """

    def __post_init__(self):
        self.lower_limits = self.lower_limits or DerivativeMap()
        self.upper_limits = self.upper_limits or DerivativeMap()
        for derivative in Derivatives.range(Derivatives.position, Derivatives.jerk):
            s = symbol_manager.register_symbol_provider(f'{self.name}_{derivative}',
                                                        lambda d=derivative: self._world.state[self.name][d])
            self.symbols.data[derivative] = s

    def reset_cache(self):
        for method_name in dir(self):
            try:
                getattr(self, method_name).memo.clear()
            except:
                pass

    @lru_cache(maxsize=None)
    def has_position_limits(self) -> bool:
        try:
            lower_limit = self.lower_limits.position
            upper_limit = self.upper_limits.position
            return lower_limit is not None or upper_limit is not None
        except KeyError:
            return False

    def __hash__(self):
        return hash(id(self))
